/*
   I, the copyright holder of this work, hereby release it into the public
   domain. This applies worldwide. In case this is not legally possible: I
   grant anyone the right to use this work for any purpose, without any
   conditions, unless such conditions are required by law.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <time.h>
#include <sys/wait.h>
#include <notcurses/direct.h>

#define CONFIG_MOVES_PER_DAY 60
#define CONFIG_MAX_INVENTORY 4

#define MAX_PLAYERS 30
#define MAX_ITEMS 91

#define MAX_PLAYER_NAME 30
#define MAX_PASSWORD 10
#define MAX_ITEM_NAME 20
#define MAX_ROOM_DESCRIPTION 55
#define MAX_CSV_ELEMENTS 10

#define BLACK 0x000000
#define DARK_GREY 0x737676
#define DARK_RED 0xcc0403
#define BRIGHT_RED 0xf2201f
#define DARK_GREEN 0x19cb00
#define BRIGHT_GREEN 0x23fd00
#define DARK_YELLOW 0xcecb00
#define BRIGHT_YELLOW 0xfffd00
#define DARK_BLUE 0x0d73cc
#define BRIGHT_BLUE 0x1a8fff
#define DARK_PURPLE 0xcb1ed1
#define BRIGHT_PURPLE 0xfd28ff
#define DARK_CYAN 0x0dcdcd
#define BRIGHT_CYAN 0x14ffff
#define GREY 0xdddddd
#define WHITE 0xffffff
#define BROWN 0xaa5500

#define ITEM_SPECIAL      (1<<0)
#define ITEM_SET_1        (1<<1)
#define ITEM_SET_2        (1<<2)
#define ITEM_SET_3        (1<<3)
#define ITEM_SET_4        (1<<4)
#define ITEM_SET_5        (1<<5)
#define ITEM_SET_6        (1<<6)
#define ITEM_SET_7        (1<<7)
#define ITEM_SET_8        (1<<8)
#define ITEM_SET_9        (1<<9)
#define ITEM_PART_A       (1<<10)
#define ITEM_PART_B       (1<<11)
#define ITEM_DEFENSE      (1<<12)
#define ITEM_IN_ROOM      (1<<13)
#define ITEM_IN_INVENTORY (1<<14)

#define PLAYER_INITIALIZED (1<<0)
#define PLAYER_ALIVE       (1<<1)
#define PLAYER_NPC         (1<<2)

#define SQUEAL_FALSE 0
#define SQUEAL_TRUE 1
#define SQUEAL_END 2

struct Room {
    unsigned short id;
    char description[MAX_ROOM_DESCRIPTION+1];
    struct Room *exits[6]; /* NSWEUD */
};

struct Item {
    unsigned short id;
    char name[MAX_ITEM_NAME+1];
    unsigned color;
    unsigned short bits;
    unsigned short location;
};

struct Player {
    unsigned short id;
    char name[MAX_PLAYER_NAME+1];
    char password[MAX_PASSWORD+1];
    struct Room *room;
    unsigned short bits;
    unsigned short moves;
    time_t moves_t;
    struct Player *target;
    unsigned short starting_room; /* 0 = rnd */
    unsigned short kills;
    unsigned short killed;
};

struct Squeal {
    time_t squeal_t;
    char name[MAX_PLAYER_NAME+1];
    char description[MAX_ROOM_DESCRIPTION+1];
    unsigned short truthful;
};

/* players.c */
struct Player * initialize_players();
void player_start(struct ncdirect *nc, struct Player *player, struct Player *players);
bool write_player_data(struct Player *players);
bool read_player_data(struct Player *players);
struct Player * match_player_name(char *str, struct Player *players);
bool check_player_password(char *str, struct Player *players);
struct Player * create_new_player(struct Player *players);
unsigned short items_in_player_inventory(struct Player *player, struct Item *items);
bool player_moves_check(struct ncdirect *nc, struct Player *players);
unsigned short count_players(struct Player *players, unsigned short bitfield);
void check_player_placement(struct Player *players, struct Room *rooms);
bool is_target(struct Player *player, struct Player *players);
struct Player * targeted_by(struct Player *player, struct Player *players);
struct Player ** find_untargeted_players(struct Player *player, struct Player *players);
struct Player ** find_alive_players(struct Player *players);
bool write_obit(struct Player *killer, struct Player *victim, char *weaponstr, struct Room *room);
struct Player * find_player_by_name(struct Player *players, char *str);

/* items.c */
struct Item * initialize_items();
void check_item_placement(struct Item *items, struct Room *rooms);
struct Item * find_item_on_player(struct Player *player, struct Item *items, unsigned short bitfield);
char * find_item_pair_on_player(struct Player *player, struct Item *items, unsigned short bitfield);
unsigned short items_in_room(struct Room *room, struct Item *items);
bool write_item_data(struct Item *items);
bool read_item_data(struct Item *items);

/* map.c */
unsigned short count_lines_in_file(const char *filename);
struct Room * initialize_rooms_from_map(const char *filename, bool empty_rooms);
unsigned short count_rooms(struct Room *rooms);
bool create_empty_map(char *filename);
bool write_map(char *filename, struct Room *rooms);
void create_exit(struct ncdirect *nc, struct Room *room);
void modify_description(struct ncdirect *nc, struct Room *room);

/* display.c */
void center_string(struct ncdirect *nc, char *str);
void display_room(struct ncdirect *nc, struct Player *player, struct Item *items, struct Player *players);
bool navigate(struct ncdirect *nc, struct Player *player, struct Item *items, struct Player *players);
unsigned short display_items_in_room(struct ncdirect *nc, struct Room *room, struct Item *items);
unsigned short display_players_in_room(struct ncdirect *nc, struct Room *room, struct Player *players, struct Player *player);
void display_navigation_help(struct ncdirect *nc);
void take_item(struct ncdirect *nc, struct Player *player, struct Item *items);
void leave_item(struct ncdirect *nc, struct Player *player, struct Item *items);
bool display_inventory(struct ncdirect *nc, struct Player *player, struct Item *items);
void display_item(struct ncdirect *nc, struct Item *item);
bool kill_player(struct ncdirect *nc, struct Player *player, struct Item *items);
unsigned short defender_has_weapon_pair(struct ncdirect *nc, struct Player *player, struct Item *items);
void most_wanted(struct ncdirect *nc, struct Player *players);
void display_obits(struct ncdirect *nc);
bool squeals_menu(struct ncdirect *nc, struct Player *player, struct Player *players);
void add_squeal(struct ncdirect *nc, struct Player *player, struct Player *players);
void read_squeals_ui(struct ncdirect *nc);
bool display_file(struct ncdirect *nc, char *filename);
void random_encounter(struct ncdirect *nc, struct Player *player);

/* squeals.c */
struct Squeal * read_squeals_file();
bool compare_squeals(struct Squeal *squeal, struct Squeal *squeals);
bool write_squeal(struct Squeal *squeal);

/* utils.c */
void uc_str(char *str);
char ** csv_line_to_null_terminated_multiple(char *str);
bool yes_no(struct ncdirect *nc, char *str);
void seed_rand();
void alphanumeric_only(char *src, char *dst);
unsigned short wrap_line(struct ncdirect *nc, char *str);
void enter_to_continue(struct ncdirect *nc, unsigned color);
void fernando_center(struct ncdirect *nc, char *str);
bool check_pid();
bool is_process_alive(pid_t pid);
